@@ -107,9 +107,9 @@ static int do_mkdir(const char *parent, const char *path, mode_t mode)
 	/* If path exists we're done */
 	status = stat(path, &st);
 	if (status == 0) {
+		errno = EEXIST;
 		if (!S_ISDIR(st.st_mode))
 			errno = ENOTDIR;
-		errno = EEXIST;
 		return 0;
 	}
 
@@ -122,7 +122,10 @@ static int do_mkdir(const char *parent, const char *path, mode_t mode)
 		status = statfs(parent, &fs);
 	if ((status != -1 && fs.f_type == (__SWORD_TYPE) AUTOFS_SUPER_MAGIC) ||
 	    contained_in_local_fs(path)) {
-		if (mkdir(path, mode) == -1) {
+		mode_t mask = umask(0022);
+		int ret = mkdir(path, mode);
+		(void) umask(mask);
+		if (ret == -1) {
 			errno = EACCES;
 			return 0;
 		}
@@ -251,10 +254,12 @@ int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev)
 /* Like ftw, except fn gets called twice: before a directory is
    entered, and after.  If the before call returns 0, the directory
    isn't entered. */
-static int walk_tree(const char *base, int (*fn) (unsigned logopt,
+static int walk_tree(const char *base, int (*fn) (struct autofs_point *ap,
 						  const char *file,
 						  const struct stat * st,
-						  int, void *), int incl, unsigned logopt, void *arg)
+						  int, void *), int incl,
+						  struct autofs_point *ap,
+						  void *arg)
 {
 	char buf[PATH_MAX + 1];
 	struct stat st, *pst = &st;
@@ -267,7 +272,7 @@ static int walk_tree(const char *base, int (*fn) (unsigned logopt,
 		ret = 0;
 	}
 
-	if (ret != -1 && (fn) (logopt, base, pst, 0, arg)) {
+	if (ret != -1 && (fn) (ap, base, pst, 0, arg)) {
 		if (S_ISDIR(st.st_mode)) {
 			struct dirent **de;
 			int n;
@@ -295,18 +300,20 @@ static int walk_tree(const char *base, int (*fn) (unsigned logopt,
 					return -1;
 				}
 
-				walk_tree(buf, fn, 1, logopt, arg);
+				walk_tree(buf, fn, 1, ap, arg);
 				free(de[n]);
 			}
 			free(de);
 		}
 		if (incl)
-			(fn) (logopt, base, pst, 1, arg);
+			(fn) (ap, base, pst, 1, arg);
 	}
 	return 0;
 }
 
-static int rm_unwanted_fn(unsigned logopt, const char *file, const struct stat *st, int when, void *arg)
+static int rm_unwanted_fn(struct autofs_point *ap,
+			  const char *file, const struct stat *st,
+			  int when, void *arg)
 {
 	dev_t dev = *(dev_t *) arg;
 	char buf[MAX_ERR_BUF];
@@ -322,38 +329,40 @@ static int rm_unwanted_fn(unsigned logopt, const char *file, const struct stat *
 	}
 
 	if (lstat(file, &newst)) {
-		crit(logopt, "unable to stat file, possible race condition");
+		crit(ap->logopt,
+		     "unable to stat file, possible race condition");
 		return 0;
 	}
 
 	if (newst.st_dev != dev) {
-		crit(logopt, "file %s has the wrong device, possible race condition",
+		crit(ap->logopt,
+		     "file %s has the wrong device, possible race condition",
 		     file);
 		return 0;
 	}
 
 	if (S_ISDIR(newst.st_mode)) {
-		debug(logopt, "removing directory %s", file);
+		debug(ap->logopt, "removing directory %s", file);
 		if (rmdir(file)) {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			warn(logopt,
+			warn(ap->logopt,
 			      "unable to remove directory %s: %s", file, estr);
 			return 0;
 		}
 	} else if (S_ISREG(newst.st_mode)) {
-		crit(logopt, "attempting to remove files from a mounted "
+		crit(ap->logopt, "attempting to remove files from a mounted "
 		     "directory. file %s", file);
 		return 0;
 	} else if (S_ISLNK(newst.st_mode)) {
-		debug(logopt, "removing symlink %s", file);
+		debug(ap->logopt, "removing symlink %s", file);
 		unlink(file);
 	}
 	return 1;
 }
 
-void rm_unwanted(unsigned logopt, const char *path, int incl, dev_t dev)
+void rm_unwanted(struct autofs_point *ap, const char *path, int incl)
 {
-	walk_tree(path, rm_unwanted_fn, incl, logopt, &dev);
+	walk_tree(path, rm_unwanted_fn, incl, ap, &ap->dev);
 }
 
 struct counter_args {
@@ -361,7 +370,8 @@ struct counter_args {
 	dev_t dev;
 };
 
-static int counter_fn(unsigned logopt, const char *file, const struct stat *st, int when, void *arg)
+static int counter_fn(struct autofs_point *ap, const char *file,
+		      const struct stat *st, int when, void *arg)
 {
 	struct counter_args *counter = (struct counter_args *) arg;
 
@@ -375,14 +385,14 @@ static int counter_fn(unsigned logopt, const char *file, const struct stat *st,
 }
 
 /* Count mounted filesystems and symlinks */
-int count_mounts(unsigned logopt, const char *path, dev_t dev)
+int count_mounts(struct autofs_point *ap, const char *path, dev_t dev)
 {
 	struct counter_args counter;
 
 	counter.count = 0;
 	counter.dev = dev;
 	
-	if (walk_tree(path, counter_fn, 0, logopt, &counter) == -1)
+	if (walk_tree(path, counter_fn, 0, ap, &counter) == -1)
 		return -1;
 
 	return counter.count;
@@ -406,9 +416,9 @@ static void check_rm_dirs(struct autofs_point *ap, const char *path, int incl)
 	    (ap->state == ST_SHUTDOWN_PENDING ||
 	     ap->state == ST_SHUTDOWN_FORCE ||
 	     ap->state == ST_SHUTDOWN))
-		rm_unwanted(ap->logopt, path, incl, ap->dev);
+		rm_unwanted(ap, path, incl);
 	else if ((ap->flags & MOUNT_FLAG_GHOST) && (ap->type == LKP_INDIRECT))
-		rm_unwanted(ap->logopt, path, 0, ap->dev);
+		rm_unwanted(ap, path, 0);
 }
 
 /* Try to purge cache entries kept around due to existing mounts */
@@ -516,7 +526,8 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 	 */
 	if (!is_mm_root && is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
 		debug(ap->logopt, "unmounting dir = %s", path);
-		if (umount_ent(ap, path)) {
+		if (umount_ent(ap, path) &&
+		    is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
 			warn(ap->logopt, "could not umount dir %s", path);
 			left++;
 		}
@@ -544,13 +555,15 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 	 * If we are a submount we need to umount any offsets our
 	 * parent may have mounted over top of us.
 	 */
-	/*if (ap->submount)
-		left += umount_subtree_mounts(ap->parent, path, is_autofs_fs);*/
+	if (ap->submount)
+		left += umount_subtree_mounts(ap->parent, path, is_autofs_fs);
 
 	left += umount_subtree_mounts(ap, path, is_autofs_fs);
 
 	/* Delete detritus like unwanted mountpoints and symlinks */
-	if (left == 0 && ap->state != ST_READMAP) {
+	if (left == 0 &&
+	    ap->state != ST_READMAP &&
+	    !count_mounts(ap, path, ap->dev)) {
 		update_map_cache(ap, path);
 		check_rm_dirs(ap, path, incl);
 	}
@@ -610,6 +623,7 @@ static size_t get_kpkt_len(void)
 		if (strcmp(un.machine, "alpha") == 0 ||
 		    strcmp(un.machine, "ia64") == 0 ||
 		    strcmp(un.machine, "x86_64") == 0 ||
+		    strcmp(un.machine, "parisc64") == 0 ||
 		    strcmp(un.machine, "ppc64") == 0)
 			pkt_len += 4;
 
@@ -1271,9 +1285,10 @@ static void *do_read_master(void *arg)
 	return NULL;
 }
 
-static int do_hup_signal(struct master *master, time_t age)
+static int do_hup_signal(struct master *master)
 {
 	unsigned int logopt = master->logopt;
+	time_t age = time(NULL);
 	pthread_t thid;
 	int status;
 
@@ -1284,6 +1299,7 @@ static int do_hup_signal(struct master *master, time_t age)
 	nfs_mount_uses_string_options = check_nfs_mount_version(&vers, &check);
 
 	master_mutex_lock();
+	/* Already doing a map read or shutdown or no mounts */
 	if (master->reading) {
 		status = pthread_mutex_unlock(&mrc.mutex);
 		if (status)
@@ -1361,7 +1377,7 @@ static void *statemachine(void *arg)
 			break;
 
 		case SIGHUP:
-			do_hup_signal(master_list, time(NULL));
+			do_hup_signal(master_list);
 			break;
 
 		default:
@@ -1448,6 +1464,7 @@ static void handle_mounts_cleanup(void *arg)
 	char path[PATH_MAX + 1];
 	char buf[MAX_ERR_BUF];
 	unsigned int clean = 0, submount, logopt;
+	unsigned int pending = 0;
 
 	ap = (struct autofs_point *) arg;
 
@@ -1465,6 +1482,9 @@ static void handle_mounts_cleanup(void *arg)
 		list_del_init(&ap->mounts);
 	}
 
+	/* Don't signal the handler if we have already done so */
+	if (!list_empty(&master_list->completed))
+		pending = 1;
 	master_remove_mapent(ap->entry);
 	master_source_unlock(ap->entry);
 
@@ -1497,7 +1517,7 @@ static void handle_mounts_cleanup(void *arg)
 	 * so it can join with any completed handle_mounts() threads and
 	 * perform final cleanup.
 	 */
-	if (!submount)
+	if (!submount && !pending)
 		pthread_kill(state_mach_thid, SIGTERM);
 
 	master_mutex_unlock();
@@ -1719,7 +1739,8 @@ static void usage(void)
 		"	-f --foreground do not fork into background\n"
 		"	-r --random-multimount-selection\n"
 		"			use ramdom replicated server selection\n"
-		"	-m --dumpmaps	dump automounter maps and exit\n"
+		"	-m --dumpmaps [<map type> <map name>]\n"
+		"			dump automounter maps and exit\n"
 		"	-n --negative-timeout n\n"
 		"			set the timeout for failed key lookups.\n"
 		"	-O --global-options\n"
@@ -1728,6 +1749,10 @@ static void usage(void)
 		"			set daemon log verbosity\n"
 		"	-C --dont-check-daemon\n"
 		"			don't check if daemon is already running\n"
+		"	-F --force	forceably clean up known automounts at start\n"
+		"	-M --master-wait n\n"
+		"			maximum time to wait for the master map to\n"
+		"			become available at start up\n"
 		"	-V --version	print version, build config and exit\n"
 		, program);
 }
@@ -1756,12 +1781,6 @@ static void show_build_info(void)
 #ifdef ENABLE_FORCED_SHUTDOWN
 	printf("ENABLE_FORCED_SHUTDOWN ");
 	count = count + 23;
-
-	if (count > 60) {
-		printf("\n  ");
-		count = 0;
-	}
-
 #endif
 
 #ifdef ENABLE_IGNORE_BUSY_MOUNTS
@@ -1898,16 +1917,61 @@ static void remove_empty_args(char **argv, int *argc)
 	*argc = j;
 }
 
+static int do_master_read_master(struct master *master, int wait)
+{
+	sigset_t signalset;
+	/* Wait must be at least 1 second */
+	unsigned int retry_wait = 2;
+	unsigned int elapsed = 0;
+	int max_wait = wait;
+	int ret = 0;
+	time_t age;
+
+	sigemptyset(&signalset);
+	sigaddset(&signalset, SIGTERM);
+	sigaddset(&signalset, SIGINT);
+	sigaddset(&signalset, SIGHUP);
+	sigprocmask(SIG_UNBLOCK, &signalset, NULL);
+
+	while (1) {
+		struct timespec t = { retry_wait, 0 };
+
+		age = time(NULL);
+		if (master_read_master(master, age, 0)) {
+			ret = 1;
+			break;
+		}
+
+		if (nanosleep(&t, NULL) == -1)
+			break;
+
+		if (max_wait > 0) {
+			elapsed += retry_wait;
+			if (elapsed >= max_wait) {
+				logmsg("problem reading master map, "
+					"maximum wait exceeded");
+				break;
+			}
+		}
+	}
+
+	sigprocmask(SIG_BLOCK, &signalset, NULL);
+
+	return ret;
+}
+
 int main(int argc, char *argv[])
 {
 	int res, opt, status;
 	int logpri = -1;
 	unsigned ghost, logging, daemon_check;
 	unsigned dumpmaps, foreground, have_global_options;
+	unsigned master_read;
+	int master_wait;
 	time_t timeout;
 	time_t age = time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:fVrO:l:n:CF";
+	const char *options = "+hp:t:vmdD:fVrO:l:n:CFM";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
@@ -1924,6 +1988,7 @@ int main(int argc, char *argv[])
 		{"set-log-priority", 1, 0, 'l'},
 		{"dont-check-daemon", 0, 0, 'C'},
 		{"force", 0, 0, 'F'},
+		{"master-wait", 1, 0, 'M'},
 		{0, 0, 0, 0}
 	};
 
@@ -1944,6 +2009,7 @@ int main(int argc, char *argv[])
 	nfs_mount_uses_string_options = check_nfs_mount_version(&vers, &check);
 
 	kpkt_len = get_kpkt_len();
+	master_wait = defaults_get_master_wait();
 	timeout = defaults_get_timeout();
 	ghost = defaults_get_browse_mode();
 	logging = defaults_get_logging();
@@ -2003,6 +2069,10 @@ int main(int argc, char *argv[])
 			dumpmaps = 1;
 			break;
 
+		case 'M':
+			master_wait = getnumopt(optarg, opt);
+			break;
+
 		case 'O':
 			if (!have_global_options) {
 				global_options = strdup(optarg);
@@ -2099,8 +2169,11 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-	rlim.rlim_cur = MAX_OPEN_FILES;
-	rlim.rlim_max = MAX_OPEN_FILES;
+	res = getrlimit(RLIMIT_NOFILE, &rlim);
+	if (res == -1 || rlim.rlim_max <= MAX_OPEN_FILES)  {
+		rlim.rlim_cur = MAX_OPEN_FILES;
+		rlim.rlim_max = MAX_OPEN_FILES;
+	}
 	res = setrlimit(RLIMIT_NOFILE, &rlim);
 	if (res)
 		printf("%s: can't increase open file limit - continuing",
@@ -2115,20 +2188,37 @@ int main(int argc, char *argv[])
 			program);
 #endif
 
-	if (argc == 0)
-		master_list = master_new(NULL, timeout, ghost);
-	else
-		master_list = master_new(argv[0], timeout, ghost);
-
-	if (!master_list) {
-		printf("%s: can't create master map %s", program, argv[0]);
-		exit(1);
-	}
+	/* Get processor information for predefined escapes */
+	macro_init();
 
 	if (dumpmaps) {
+		struct master_mapent *entry;
+		struct list_head *head, *p;
 		struct mapent_cache *nc;
+		const char *type = NULL;
+		const char *name = NULL;
+		const char *master = NULL;
 
-		open_log();
+		if (argc > 0) {
+			if (argc >= 2) {
+				type = argv[0];
+				name = argv[1];
+			}
+			if (argc == 3)
+				master = argv[2];
+		}
+
+		if (master)
+			master_list = master_new(NULL, timeout, ghost);
+		else
+			master_list = master_new(master, timeout, ghost);
+		if (!master_list) {
+			printf("%s: can't create master map", program);
+			macro_free_global_table();
+			exit(1);
+		}
+
+		log_to_stderr();
 
 		master_init_scan();
 
@@ -2136,15 +2226,47 @@ int main(int argc, char *argv[])
 		if (!nc) {
 			printf("%s: failed to init null map cache for %s",
 				program, master_list->name);
+			macro_free_global_table();
 			exit(1);
 		}
 		master_list->nc = nc;
 
 		lookup_nss_read_master(master_list, 0);
-		master_show_mounts(master_list);
+		if (type) {
+			const char *map = basename(name);
+			if (!map)
+				printf("%s: invalid map name %s\n",
+					program, name);
+			else
+				dump_map(master_list, type, map);
+		} else
+			master_show_mounts(master_list);
+
+		head = &master_list->mounts;
+		p = head->next;
+		while (p != head) {
+			entry = list_entry(p, struct master_mapent, list);
+			p = p->next;
+			master_free_mapent_sources(entry, 1);
+			master_free_mapent(entry);
+		}
+		master_kill(master_list);
+		macro_free_global_table();
+
 		exit(0);
 	}
 
+	if (argc == 0)
+		master_list = master_new(NULL, timeout, ghost);
+	else
+		master_list = master_new(argv[0], timeout, ghost);
+
+	if (!master_list) {
+		printf("%s: can't create master map %s", program, argv[0]);
+		macro_free_global_table();
+		exit(1);
+	}
+
 	become_daemon(foreground, daemon_check);
 
 	if (pthread_attr_init(&th_attr)) {
@@ -2153,6 +2275,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2162,6 +2285,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2172,6 +2296,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2183,6 +2308,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 #endif
@@ -2201,6 +2327,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2212,6 +2339,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2221,6 +2349,7 @@ int main(int argc, char *argv[])
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 		close(start_pipefd[1]);
 		release_flag_file();
+		macro_free_global_table();
 		exit(1);
 	}
 
@@ -2237,13 +2366,16 @@ int main(int argc, char *argv[])
 		dh_tirpc = dlopen("libtirpc.so.1", RTLD_NOW);
 #endif
 
-	if (!master_read_master(master_list, age, 0)) {
-		master_kill(master_list);
-		*pst_stat = 3;
-		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
-		close(start_pipefd[1]);
-		release_flag_file();
-		exit(3);
+	master_read = master_read_master(master_list, age, 0);
+	if (!master_read) {
+		if (foreground)
+			logerr("%s: failed to read master map, "
+			       "will retry!",
+			       program);
+		else
+			logerr("%s: failed to read master map, "
+			       "will retry in background!",
+			       program);
 	}
 
 	/*
@@ -2256,6 +2388,20 @@ int main(int argc, char *argv[])
 	res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 	close(start_pipefd[1]);
 
+	if (!master_read) {
+		/*
+		 * Read master map, waiting until it is available, unless
+		 * a signal is received, in which case exit returning an
+		 * error.
+		 */
+		if (!do_master_read_master(master_list, -1)) {
+			logerr("%s: failed to read master map!", program);
+			master_kill(master_list);
+			release_flag_file();
+			exit(3);
+		}
+	}
+
 	state_mach_thid = pthread_self();
 	statemachine(NULL);
 
@@ -2267,6 +2413,7 @@ int main(int argc, char *argv[])
 	}
 	closelog();
 	release_flag_file();
+	macro_free_global_table();
 
 #ifdef TIRPC_WORKAROUND
 	if (dh_tirpc)
