@@ -100,6 +100,27 @@ static int do_read_master(struct master *master, char *type, time_t age)
 	return status;
 }
 
+static char *find_map_path(struct map_source *map)
+{
+	struct stat st;
+	char *path;
+
+	path = malloc(strlen(AUTOFS_MAP_DIR) + strlen(map->argv[0]) + 2);
+	if (!path)
+		return NULL;
+
+	strcpy(path, AUTOFS_MAP_DIR);
+	strcat(path, "/");
+	strcat(path, map->argv[0]);
+
+	if (!stat(path, &st))
+		return path;
+
+	free(path);
+
+	return NULL;
+}
+
 static int read_master_map(struct master *master, char *type, time_t age)
 {
 	unsigned int logopt = master->logopt;
@@ -171,14 +192,22 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			char source[10];
 
 			memset(source, 0, 10);
-			if (!strncmp(name, "file:", 5) ||
-			    !strncmp(name, "yp:", 3) ||
-			    !strncmp(name, "nis:", 4) ||
-			    !strncmp(name, "nisplus:", 8) ||
-			    !strncmp(name, "ldap:", 5) ||
-			    !strncmp(name, "ldaps:", 6) ||
-			    !strncmp(name, "sss:", 4) ||
-			    !strncmp(name, "dir:", 4)) {
+			if ((!strncmp(name, "file", 4) &&
+				 (name[4] == ',' || name[4] == ':')) ||
+			    (!strncmp(name, "yp", 2) &&
+				 (name[2] == ',' || name[2] == ':')) ||
+			    (!strncmp(name, "nis", 3) &&
+				 (name[3] == ',' || name[3] == ':')) ||
+			    (!strncmp(name, "nisplus", 7) &&
+				 (name[7] == ',' || name[7] == ':')) ||
+			    (!strncmp(name, "ldap", 4) &&
+				 (name[4] == ',' || name[4] == ':')) ||
+			    (!strncmp(name, "ldaps", 5) &&
+				 (name[5] == ',' || name[5] == ':')) ||
+			    (!strncmp(name, "sss", 3) ||
+				 (name[3] == ',' || name[3] == ':')) ||
+			    (!strncmp(name, "dir", 3) &&
+				 (name[3] == ',' || name[3] == ':'))) {
 				strncpy(source, name, tmp - name);
 
 				/*
@@ -218,6 +247,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 	}
 
 	/* First one gets it */
+	result = NSS_STATUS_UNKNOWN;
 	head = &nsslist;
 	list_for_each(p, head) {
 		struct nss_source *this;
@@ -225,6 +255,12 @@ int lookup_nss_read_master(struct master *master, time_t age)
 
 		this = list_entry(p, struct nss_source, list);
 
+		if (strncmp(this->source, "files", 5) &&
+		    strncmp(this->source, "nis", 3) &&
+		    strncmp(this->source, "nisplus", 7) &&
+		    strncmp(this->source, "ldap", 4))
+			continue;
+
 		debug(logopt,
 		      "reading master %s %s", this->source, master->name);
 
@@ -384,6 +420,45 @@ static void argv_cleanup(void *arg)
 	return;
 }
 
+static int lookup_map_read_map(struct autofs_point *ap,
+			       struct map_source *map, time_t age)
+{
+	char *path;
+
+	if (!map->argv[0]) {
+		if (!strcmp(map->type, "hosts"))
+			return do_read_map(ap, map, age);
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	/*
+	 * This is only called when map->type != NULL.
+	 * We only need to look for a map if source type is
+	 * file and the map name doesn't begin with a "/".
+	 */
+	if (strncmp(map->type, "file", 4))
+		return do_read_map(ap, map, age);
+
+	if (map->argv[0][0] == '/')
+		return do_read_map(ap, map, age);
+
+	path = find_map_path(map);
+	if (!path)
+		return NSS_STATUS_UNKNOWN;
+
+	if (map->argc >= 1) {
+		if (map->argv[0])
+			free((char *) map->argv[0]);
+		map->argv[0] = path;
+	} else {
+		error(ap->logopt, "invalid arguments for autofs_point");
+		free(path);
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	return do_read_map(ap, map, age);
+}
+
 static enum nsswitch_status read_map_source(struct nss_source *this,
 		struct autofs_point *ap, struct map_source *map, time_t age)
 {
@@ -420,14 +495,10 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
-	path = malloc(strlen(AUTOFS_MAP_DIR) + strlen(map->argv[0]) + 2);
+	path = find_map_path(map);
 	if (!path)
 		return NSS_STATUS_UNKNOWN;
 
-	strcpy(path, AUTOFS_MAP_DIR);
-	strcat(path, "/");
-	strcat(path, map->argv[0]);
-
 	if (map->argc >= 1) {
 		tmap.argc = map->argc;
 		tmap.argv = copy_argv(map->argc, map->argv);
@@ -488,7 +559,7 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 				debug(ap->logopt,
 				      "reading map %s %s",
 				       map->type, map->argv[0]);
-			result = do_read_map(ap, map, age);
+			result = lookup_map_read_map(ap, map, age);
 			map = map->next;
 			continue;
 		}
@@ -604,6 +675,17 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 		cache_readlock(mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
+			/*
+			 * Map entries that have been created in the cache
+			 * due to a negative lookup but don't exist in the
+			 * map source shouldn't have directories created.
+			 * me->status of negative entries that are present
+			 * in the map source will have me->status set to 0
+			 * in lookup_prune_one_cache().
+			 */
+			if (me->status && !me->mapent)
+				goto next;
+
 			if (!strcmp(me->key, "*"))
 				goto next;
 
@@ -742,6 +824,46 @@ static int lookup_name_source_instance(struct autofs_point *ap, struct map_sourc
 	return do_lookup_mount(ap, instance, name, name_len);
 }
 
+static int do_name_lookup_mount(struct autofs_point *ap,
+				struct map_source *map,
+				const char *name, int name_len)
+{
+	char *path;
+
+	if (!map->argv[0]) {
+		if (!strcmp(map->type, "hosts"))
+			return do_lookup_mount(ap, map, name, name_len);
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	/*
+	 * This is only called when map->type != NULL.
+	 * We only need to look for a map if source type is
+	 * file and the map name doesn't begin with a "/".
+	 */
+	if (strncmp(map->type, "file", 4))
+		return do_lookup_mount(ap, map, name, name_len);
+
+	if (map->argv[0][0] == '/')
+		return do_lookup_mount(ap, map, name, name_len);
+
+	path = find_map_path(map);
+	if (!path)
+		return NSS_STATUS_UNKNOWN;
+
+	if (map->argc >= 1) {
+		if (map->argv[0])
+			free((char *) map->argv[0]);
+		map->argv[0] = path;
+	} else {
+		error(ap->logopt, "invalid arguments for autofs_point");
+		free(path);
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	return do_lookup_mount(ap, map, name, name_len);
+}
+
 static enum nsswitch_status lookup_map_name(struct nss_source *this,
 			struct autofs_point *ap, struct map_source *map,
 			const char *name, int name_len)
@@ -777,14 +899,10 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 	tmap.argc = 0;
 	tmap.argv = NULL;
 
-	path = malloc(strlen(AUTOFS_MAP_DIR) + strlen(map->argv[0]) + 2);
+	path = find_map_path(map);
 	if (!path)
 		return NSS_STATUS_UNKNOWN;
 
-	strcpy(path, AUTOFS_MAP_DIR);
-	strcat(path, "/");
-	strcat(path, map->argv[0]);
-
 	if (map->argc >= 1) {
 		tmap.argc = map->argc;
 		tmap.argv = copy_argv(map->argc, map->argv);
@@ -822,6 +940,10 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 	if (source && source->depth)
 		return;
 
+	/* Don't update the wildcard */
+	if (strlen(name) == 1 && *name == '*')
+		return;
+
 	/* Have we recorded the lookup fail for negative caching? */
 	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
 	if (me)
@@ -844,10 +966,15 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 			int rv = CHE_FAIL;
 
 			cache_writelock(map->mc);
-			rv = cache_update(map->mc, map, name, NULL, now);
+			me = cache_lookup_distinct(map->mc, name);
+			if (me)
+				rv = cache_push_mapent(me, NULL);
+			else
+				rv = cache_update(map->mc, map, name, NULL, now);
 			if (rv != CHE_FAIL) {
 				me = cache_lookup_distinct(map->mc, name);
-				me->status = now + ap->negative_timeout;
+				if (me)
+					me->status = now + ap->negative_timeout;
 			}
 			cache_unlock(map->mc);
 		}
@@ -889,8 +1016,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		sched_yield();
 
 		if (map->type) {
-			result = do_lookup_mount(ap, map, name, name_len);
-
+			result = do_name_lookup_mount(ap, map, name, name_len);
 			if (result == NSS_STATUS_SUCCESS)
 				break;
 
@@ -934,10 +1060,8 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 
 			result = lookup_map_name(this, ap, map, name, name_len);
 
-			if (result == NSS_STATUS_UNKNOWN) {
-				map = map->next;
+			if (result == NSS_STATUS_UNKNOWN)
 				continue;
-			}
 
 			status = check_nss_result(this, result);
 			if (status >= 0) {
@@ -1042,7 +1166,8 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 			 * Reset time of last fail for valid map entries to
 			 * force entry update and subsequent mount retry.
 			 */
-			me->status = 0;
+			if (me->mapent)
+				me->status = 0;
 			me = cache_enumerate(mc, me);
 			continue;
 		}
@@ -1102,7 +1227,10 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 
 		if (valid)
 			cache_delete(mc, key);
-		else if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
+		else if (this->status) {
+			cache_unlock(mc);
+			goto next;
+		 } else if (!is_mounted(_PROC_MOUNTS, path, MNTS_AUTOFS)) {
 			dev_t devid = ap->dev;
 			status = CHE_FAIL;
 			if (ap->type == LKP_DIRECT)
