@@ -221,7 +221,7 @@ const char **copy_argv(int argc, const char **argv);
 int compare_argv(int argc1, const char **argv1, int argc2, const char **argv2);
 int free_argv(int argc, const char **argv);
 
-inline void dump_core(void);
+void dump_core(void);
 int aquire_lock(void);
 void release_lock(void);
 int spawnl(unsigned logopt, const char *prog, ...);
